/*
 *	libnetfilter_xtables
 *	Copyright © Jan Engelhardt, 2010-2012
 *
 *	This program is free software: you can redistribute it and/or
 *	modify it under the terms of the GNU Lesser General Public License
 *	as published by the Free Software Foundation, either version 2.1 of
 *	the License, or (at your option) any later version.
 */
#include <errno.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <pthread.h>
#include <libHX/defs.h>
#include <libHX/init.h>
#include <libHX/map.h>
#include <libnetfilter_xtables/xtuser.h>
#include <linux/netfilter/nfnetlink_xtables.h>
#include "internal.h"

static pthread_mutex_t xtuser_init_lock = PTHREAD_MUTEX_INITIALIZER;
static unsigned long xtuser_use_count;

static struct HXmap *xtuser_estr2int_map, *xtuser_eint2str_map;

/**
 * Initialize the library. This is refcounted and properly nests, and all
 * libnetfilter_xtables users (including other shared libraries that make use
 * of our functions) must call this.
 */
EXPORT_SYMBOL int xtuser_init(void)
{
	int ret;

	pthread_mutex_lock(&xtuser_init_lock);
	if (xtuser_use_count == 0) {
		ret = HX_init();
		if (ret <= 0)
			goto out;
		xtuser_estr2int_map = HXmap_init(HXMAPT_DEFAULT, HXMAP_SCKEY);
		if (xtuser_estr2int_map == NULL)
			goto err;
		xtuser_eint2str_map = HXmap_init(HXMAPT_DEFAULT, HXMAP_SDATA);
		if (xtuser_eint2str_map == NULL)
			goto err;
	}
	++xtuser_use_count;
	pthread_mutex_unlock(&xtuser_init_lock);
	return 0;

 err:
	ret = -errno;
	HX_exit();
	if (xtuser_estr2int_map != NULL)
		HXmap_free(xtuser_estr2int_map);
 out:
	return ret;
}

/**
 * Drop the library reference count. Will deinitialize all the library-specific
 * internal data structures once possible.
 */
EXPORT_SYMBOL void xtuser_exit(void)
{
	pthread_mutex_lock(&xtuser_init_lock);
	if (xtuser_use_count == 0) {
		fprintf(stderr, "%s: reference count is already zero!\n",
		        __func__);
	} else if (xtuser_use_count == 1) {
		--xtuser_use_count;
		HXmap_free(xtuser_estr2int_map);
		HXmap_free(xtuser_eint2str_map);
		HX_exit();
	} else {
		--xtuser_use_count;
	}
	pthread_mutex_unlock(&xtuser_init_lock);
}

/**
 * Turn an error code as returned by the other nfxt functions into a
 * readable string.
 */
EXPORT_SYMBOL const char *xtuser_strerror(int code)
{
	const char *ret;

	switch (code) {
	case -EINVAL:
		return "Invalid argument: The Netlink message or attribute "
		       "type was not recognized, or an attribute failed to "
		       "validate. That, or see dmesg.";
	case -EMSGSIZE:
		return "Limitations of Netlink in the kernel";
	}
	if (code <= 0)
		return strerror(-code);
	switch (code) {
	case NFXTE_ATTRSET_INCOMPLETE:
		return "Required attributes were absent from query";
	case NFXTE_CHAIN_INVALID_NAME:
		return "The name is not acceptable";
	case NFXTE_CHAIN_EXISTS:
		return "Chain already exists";
	case NFXTE_CHAIN_NOENT:
		return "No such chain exists";
	case NFXTE_CHAIN_NAMETOOLONG:
		return "Chain name is too long";
	case NFXTE_TRANSACT_ACTIVE:
		return "Transaction has already been started";
	case NFXTE_TRANSACT_INACTIVE:
		return "No transaction had been started";
	case NFXTE_RULE_OUT_OF_RANGE:
		return "Rule indices are out of range";
	case NFXTE_CHAIN_IS_PROMOTED:
		return "Chain is already promoted";
	case NFXTE_CHAIN_IS_DEMOTED:
		return "Chain is already demoted";
	case NFXTE_UNKNOWN_ATTRIBUTE:
		return "Netlink message contained an unrecognized "
		       "and non-ignored attribute";
	case NFXTE_ACTION_NOENT:
		return "The requested match/target extension was not found";
	case NFXTE_ACTION_MISMATCH:
		return "The extension's parameter block ABI was not honored";
	case NFXTE_ACTION_REJECTED:
		return "The extension's checkentry function rejected the rule";
	case NFXTE_RULESET_LOOP:
		return "I'm going in circles...";
	case NFXTE_JUMP_DST_NOENT:
		return "Jump/Goto target not found";
	/* Local */
	case NFXTE_EMSGSIZE:
		return "The userspace side message buffer was not big enough";
	default:
		fprintf(stderr, "%s: unknown error code %d\n", __func__, code);
		return "Unknown XTNL error code";
	}
	return ret;
}
