/*
 *	libnetfilter_xtables
 *	nfnetlink connector
 *	Copyright © Jan Engelhardt, 2010-2012
 *
 *	This program is free software: you can redistribute it and/or
 *	modify it under the terms of the GNU Lesser General Public License
 *	as published by the Free Software Foundation, either version 2.1 of
 *	the License, or (at your option) any later version.
 */
#include <sys/socket.h>
#include <sys/types.h>
#include <errno.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <libHX/defs.h>
#include <libHX/misc.h>
#include <libHX/string.h>
#include <libmnl/libmnl.h>
#include <libnetfilter_xtables/xtnetlink.h>
#include <libnetfilter_xtables/xtuser.h>
#include <netinet/in.h>
#include <linux/netlink.h>
#include <linux/netfilter.h>
#include <linux/netfilter/nfnetlink.h>
#include <linux/netfilter/nfnetlink_xtables.h>
#include "internal.h"

/**
 * @mnl_sock:	mnl socket instance
 * @buf:	buffer for mnl socket transactions
 * @hexdump:	do hexdump of all sent/received datagrams
 */
struct xtnetlink_handle {
	struct mnl_socket *mnl_sock;
	void *buf;
	size_t bufsize;
	FILE *hexdump;
	uint32_t seq;
};

/**
 * @table:	current table being processed
 * @chain:	current chain being processed
 * @name:	name of chain that is going to be requested
 * @level:	parser state: %NFXTL_*
 * @op:		Netlink request type
 *
 * State of NFXTM_{CHAIN,TABLE}_DUMP parsing.
 */
struct xtnetlink_dp_state {
	struct xtuser_table *table;
	struct xtuser_chain *chain;
	const char *name;
	enum nfxt_msg_type op;
};

typedef int (*mrecv_func_t)(const struct nlmsghdr *, void *);

static const char *xtnetlink_msg_name(unsigned int type)
{
	#define E(s) [s] = #s
	static const char *const nlmsg_data[] = {
		E(NLMSG_NOOP),
		E(NLMSG_ERROR),
		E(NLMSG_DONE),
		E(NLMSG_OVERRUN),
	};
	static const char *const nfxtm_data[] = {
		E(NFXTM_IDENTIFY),
		E(NFXTM_ERROR),
		E(NFXTM_CHAIN_CREATE),
		E(NFXTM_CHAIN_DELETE),
		E(NFXTM_COMMIT),
		E(NFXTM_TABLE_REPLACE),
		E(NFXTM_ABORT),
		E(NFXTM_CHAIN_DUMP),
		E(NFXTM_TABLE_DUMP),
		E(NFXTM_CHAIN_SPLICE),
		E(NFXTM_RULE_ENTRY),
		E(NFXTM_CHAIN_PROMOTE),
		E(NFXTM_CHAIN_DEMOTE),
	#undef E
	};
	unsigned int subsys = NFNL_SUBSYS_ID(type);

	type = NFNL_MSG_TYPE(type);
	if (subsys == NFNL_SUBSYS_NONE) {
		if (type >= ARRAY_SIZE(nlmsg_data))
			return "<out-of-range>";
		return nlmsg_data[type];
	}
	if (subsys != NFNL_SUBSYS_XTABLES)
		return "<unknown-subsys>";
	if (type >= ARRAY_SIZE(nfxtm_data))
		return "<NFXTM-unknown>";
	return nfxtm_data[type];
}

/**
 * Create a MNL/Netlink socket and connect to the Xtables2 kernel subsystem.
 * An opaque nfxt_handle is returned, or %NULL on error, with errno being set.
 */
EXPORT_SYMBOL struct xtnetlink_handle *xtnetlink_open(unsigned int flags)
{
	struct xtnetlink_handle *sk;
	const char *s;
	int saved_errno;

	sk = calloc(1, sizeof(*sk));
	if (sk == NULL)
		return NULL;
	sk->bufsize = 64 << 10;
	sk->buf = malloc(sk->bufsize);
	if (sk->buf == NULL)
		goto out;
	sk->mnl_sock = mnl_socket_open(NETLINK_NETFILTER);
	if (sk->mnl_sock == NULL)
		goto out;
	s = getenv("LIBNFXT_DEBUG");
	if (s != NULL && strtoul(s, NULL, 0) > 0)
		sk->hexdump = stdout;
	sk->seq = time(NULL);
	return sk;

 out:
	saved_errno = errno;
	free(sk->buf);
	free(sk);
	errno = saved_errno;
	return NULL;
}

/**
 * Close the Netlink socket and release all memory belonging to our handle.
 */
EXPORT_SYMBOL int xtnetlink_close(struct xtnetlink_handle *sk)
{
	int ret;

	ret = mnl_socket_close(sk->mnl_sock);
	free(sk->buf);
	free(sk);
	return ret;
}

static struct nlmsghdr *__xtnetlink_newpkt(char *buf, uint8_t type)
{
	struct nlmsghdr *nlmsg;
	struct nfgenmsg *nfmsg;

	nlmsg              = mnl_nlmsg_put_header(buf);
	nlmsg->nlmsg_type  = (NFNL_SUBSYS_XTABLES << 8) | type;
	/*
	 * Not using NLM_F_ACK, because we are not interested in af_netlink.c's
	 * low-informational acks/errs. xt2_nfnl generates its own responses.
	 */
	nlmsg->nlmsg_flags = NLM_F_REQUEST;
	nlmsg->nlmsg_seq   = 0;

	nfmsg = mnl_nlmsg_put_extra_header(nlmsg, sizeof(*nfmsg));
	nfmsg->nfgen_family = NFPROTO_UNSPEC;
	nfmsg->version      = NFNETLINK_V0;
	nfmsg->res_id       = 0;
	return nlmsg;
}

static struct nlmsghdr *
xtnetlink_newpkt(struct xtnetlink_handle *sk, uint8_t type)
{
	struct nlmsghdr *nlmsg = __xtnetlink_newpkt(sk->buf, type);
	nlmsg->nlmsg_seq = ++sk->seq;
	return nlmsg;
}

/**
 * Verify that a packet is indeed an ACK/ERROR packet, and return its error
 * code, which may be negative (for errno codes) or positive (%NFXTE_*).
 * If the packet is not an %NFXTM_ERROR message, -EIO is returned.
 */
static int xtnetlink_pkt_error(const void *buf, unsigned int len)
{
	const struct nlmsghdr *nlmsg = buf;
	const struct nlmsgerr *errmsg;
	const struct nlattr *attr;

	if (nlmsg->nlmsg_type == NLMSG_ERROR) {
		if (nlmsg->nlmsg_len < sizeof(*nlmsg) + sizeof(*errmsg))
			return -EIO;
		errmsg = mnl_nlmsg_get_payload(nlmsg);
		return (errmsg->error < 0) ? errmsg->error : -errmsg->error;
	}
	if (NFNL_SUBSYS_ID(nlmsg->nlmsg_type) != NFNL_SUBSYS_XTABLES)
		return -EIO;
	if (NFNL_MSG_TYPE(nlmsg->nlmsg_type) != NFXTM_ERROR)
		return -EIO;

	mnl_attr_for_each(attr, nlmsg, sizeof(struct nfgenmsg)) {
		unsigned int type = mnl_attr_get_type(attr);

		if (type == NFXTA_XTERRNO)
			return mnl_attr_get_u16(attr);
		else if (type == NFXTA_ERRNO)
			return -static_cast(int16_t, mnl_attr_get_u16(attr));
	}
	return -EIO;
}

/**
 * Send function with integrated hexdump.
 */
static ssize_t
xtnetlink_send(struct xtnetlink_handle *sk, const void *buf, size_t size)
{
	ssize_t ret;

	if (sk->hexdump != NULL) {
		const struct nlmsghdr *nlmsg = buf;
		fprintf(sk->hexdump, ">>> Writing %zu bytes "
		        "(len=%u, type=%s(%u) flags=0x%x)\n",
		        size, nlmsg->nlmsg_len,
			xtnetlink_msg_name(nlmsg->nlmsg_type),
			NFNL_MSG_TYPE(nlmsg->nlmsg_type), nlmsg->nlmsg_flags);
		HX_hexdump(sk->hexdump, buf, size);
	}
	ret = mnl_socket_sendto(sk->mnl_sock, buf, size);
	return (ret < 0) ? -errno : ret;
}

/**
 * Receive function with integrated hexdump.
 */
static ssize_t
xtnetlink_recv(struct xtnetlink_handle *sk, void *buf, size_t size)
{
	ssize_t ret = mnl_socket_recvfrom(sk->mnl_sock, buf, size);

	if (ret < 0)
		return -errno;
	if (sk->hexdump == NULL)
		return ret;
	if (ret >= sizeof(struct nlmsghdr)) {
		const struct nlmsghdr *nlmsg = buf;
		fprintf(sk->hexdump, "<<< Read %zd bytes "
		        "(len=%u type=%s(%u) flags=0x%x)\n",
		        ret, nlmsg->nlmsg_len,
			xtnetlink_msg_name(nlmsg->nlmsg_type),
			NFNL_MSG_TYPE(nlmsg->nlmsg_type), nlmsg->nlmsg_flags);
	} else {
		fprintf(sk->hexdump, "<<< Read %zd bytes\n", ret);
	}
	HX_hexdump(sk->hexdump, buf, ret);
	return ret;
}

/**
 * Indicate whether this Netlink message is the last in a series
 * during e.g. a dump.
 */
static inline bool xtnetlink_meof(const struct nlmsghdr *msg)
{
	return !(msg->nlmsg_flags & NLM_F_MULTI) ||
	       msg->nlmsg_type == NLMSG_DONE;
}

static ssize_t
xtnetlink_mrecv(struct xtnetlink_handle *sk, mrecv_func_t cb, void *priv)
{
	struct nlmsghdr *msg = sk->buf;
	bool have_err = false;
	ssize_t ret;
	int err = 0;

	do {
		ret = xtnetlink_recv(sk, msg, sk->bufsize);
		if (ret < 0)
			return ret;
		else if (ret < sizeof(struct nlmsghdr))
			return -EIO;
		else if (msg->nlmsg_type == NLMSG_ERROR)
			return xtnetlink_pkt_error(msg, ret);
		else if (msg->nlmsg_type == NLMSG_DONE)
			return 0;
		else if (ret < sizeof(struct nlmsghdr) +
		    sizeof(struct nfgenmsg))
			return -EIO;

		if (have_err) {
			
		} else if (NFNL_MSG_TYPE(msg->nlmsg_type) == NFXTM_ERROR) {
			have_err = true;
			err = xtnetlink_pkt_error(msg, ret);
		} else if (cb != NULL) {
			(*cb)(msg, priv);
		}
	} while (!xtnetlink_meof(msg));

	return have_err ? err : ret;
}

EXPORT_SYMBOL int xtnetlink_chain_create(struct xtnetlink_handle *sk,
                                         const struct xtuser_chain *chain)
{
	struct nlmsghdr *msg;
	int ret;

	msg = xtnetlink_newpkt(sk, NFXTM_CHAIN_CREATE);
	if (!mnl_attr_put_strz_check(msg, sk->bufsize,
	                             NFXTA_CHAIN_NAME, chain->name))
		return NFXTE_EMSGSIZE;
	ret = xtnetlink_send(sk, msg, msg->nlmsg_len);
	if (ret < 0)
		return ret;
	ret = xtnetlink_recv(sk, sk->buf, sk->bufsize);
	if (ret < 0)
		return ret;
	return xtnetlink_pkt_error(sk->buf, ret);
}

EXPORT_SYMBOL int
xtnetlink_chain_del(struct xtnetlink_handle *sk, const char *name)
{
	struct nlmsghdr *msg;
	int ret;

	msg = xtnetlink_newpkt(sk, NFXTM_CHAIN_DELETE);
	if (!mnl_attr_put_strz_check(msg, sk->bufsize, NFXTA_CHAIN_NAME, name))
		return NFXTE_EMSGSIZE;
	ret = xtnetlink_send(sk, msg, msg->nlmsg_len);
	if (ret < 0)
		return ret;
	ret = xtnetlink_recv(sk, sk->buf, sk->bufsize);
	if (ret < 0)
		return ret;
	return xtnetlink_pkt_error(sk->buf, ret);
}

#if 0
EXPORT_SYMBOL int
xtnetlink_chain_move(struct xtnetlink_handle *sk, const char *old_name,
                     const char *new_name)
{
	struct nlmsghdr *msg;
	int ret;

	msg = xtnetlink_newpkt(sk, NFXTM_CHAIN_MOVE);
	if (!mnl_attr_put_strz_check(msg, sk->bufsize,
	                             NFXTA_CHAIN_NAME, old_name) ||
	    !mnl_attr_put_strz_check(msg, sk->bufsize,
	                             NFXTA_NEW_NAME, new_name))
		return NFXTE_EMSGSIZE;
	ret = xtnetlink_send(sk, msg, msg->nlmsg_len);
	if (ret < 0)
		return ret;
	ret = xtnetlink_recv(sk, sk->buf, sk->bufsize);
	if (ret < 0)
		return ret;
	return xtnetlink_pkt_error(sk->buf, ret);
}
#endif

static int
xtnetlink_emit_verdict(struct xtnetlink_handle *sk, struct nlmsghdr *msg,
                       const struct xtuser_action *action)
{
	return !mnl_attr_put_u32_check(msg, sk->bufsize, NFXTA_VERDICT,
	       action->verdict) ? NFXTE_EMSGSIZE : 0;
}

/**
 * @sk:		xtnl handle, used for the buffer size only
 * @msg:	output Netlink message
 * @action:	input action structure
 *
 * Convert the local match/target action to Netlink attributes appended to
 * the given message.
 */
static int xtnetlink_emit_mt(struct xtnetlink_handle *sk, struct nlmsghdr *msg,
                             const struct xtuser_action *action)
{
	struct nlattr *attr;
	unsigned int attr_type;

	attr_type = (action->type == XTUSER_ACTION_MATCH) ?
	            NFXTA_MATCH : NFXTA_TARGET;
	attr = mnl_attr_nest_start_check(msg, sk->bufsize, attr_type);
	if (attr == NULL)
		return NFXTE_EMSGSIZE;
	if (!mnl_attr_put_strz_check(msg, sk->bufsize,
	    NFXTA_ACTION_NAME, action->mt.name) ||
	    !mnl_attr_put_u32_check(msg, sk->bufsize,
	    NFXTA_REVISION, action->mt.revision))
		goto out;
	if (action->mt.data != NULL && action->mt.dsize > 0 &&
	    !mnl_attr_put_check(msg, sk->bufsize,
	    NFXTA_ACTION_DATA, action->mt.dsize, action->mt.data))
		goto out;
	mnl_attr_nest_end(msg, attr);
	return 0;

 out:
	mnl_attr_nest_cancel(msg, attr);
	return NFXTE_EMSGSIZE;
}

/**
 * @sk:		xtnl handle, used for the buffer size only
 * @msg:	output Netlink message
 * @action:	input action structure
 *
 * Convert the local jump/goto action to Netlink attributes appended to the
 * given message.
 */
static int xtnetlink_emit_jg(struct xtnetlink_handle *sk, struct nlmsghdr *msg,
                             const struct xtuser_action *action)
{
	unsigned int attr_type;

	attr_type = (action->type == XTUSER_ACTION_JUMP) ?
	            NFXTA_JUMP_CHAIN : NFXTA_GOTO_CHAIN;
	if (!mnl_attr_put_strz_check(msg, sk->bufsize,
	    attr_type, action->jg_target))
		return NFXTE_EMSGSIZE;
	return 0;
}

/**
 * Emit a NFXTM_RULE_ENTRY Netlink message for the given local rule.
 */
static int xtnetlink_emit_rule(struct xtnetlink_handle *sk,
                               const struct xtuser_rule *rule)
{
	const struct xtuser_action *action;
	struct nlmsghdr *msg;
	int ret;

	msg = xtnetlink_newpkt(sk, NFXTM_RULE_ENTRY);

	HXlist_for_each_entry(action, &rule->action_list, anchor) {
		if (action->type == XTUSER_ACTION_VERDICT) {
			ret = xtnetlink_emit_verdict(sk, msg, action);
		} else if (action->type == XTUSER_ACTION_MATCH ||
		    action->type == XTUSER_ACTION_TARGET) {
			ret = xtnetlink_emit_mt(sk, msg, action);
		} else if (action->type == XTUSER_ACTION_JUMP ||
		    action->type == XTUSER_ACTION_GOTO) {
			ret = xtnetlink_emit_jg(sk, msg, action);
		} else {
			fprintf(stderr, "WARNING: %s: Unknown type %u\n",
			        __func__, action->type);
			ret = -EINVAL;
		}
		if (ret != 0)
			return ret;
	}

	ret = xtnetlink_send(sk, msg, msg->nlmsg_len);
	if (ret < 0)
		return ret;
	ret = xtnetlink_recv(sk, sk->buf, sk->bufsize);
	if (ret < 0)
		return ret;
	return xtnetlink_pkt_error(sk->buf, ret);
}

/**
 * @sk:		xtnetlink handle
 * @chain:	list of rules to insert; also contains name of chain to modify
 * @offset:	rule index (0-based) where to delete/insert rules
 * @dlength:	number of rules to delete
 *
 * Does a splice operation (like Perl's similar-named function), a combined
 * delete-and-insert. The chain structure merely serves to carry around the
 * rules and the name.
 *
 * @offset may be %XTNL_CHAIN_APPEND, in which case @length is ignored.
 * @dlength may be %XTNL_CHAIN_FLUSH, in which case @offset should be 0.
 *
 * Examples:
 * Append at end:
 * 	xtnetlink_chain_splice(sk, c, XTNL_CHAIN_APPEND, 0)
 * Insert at index:
 * 	xtnetlink_chain_splice(sk, c, offset, 0)
 * Replace (one rule by rules in c) at index:
 * 	xtnetlink_chain_splice(sk, c, offset, 1)
 * Flush chain:
 * 	xtnetlink_chain_splice(sk, c, 0, XTNL_CHAIN_FLUSH)
 */
EXPORT_SYMBOL int
xtnetlink_chain_splice(struct xtnetlink_handle *sk,
                       const struct xtuser_chain *chain,
                       unsigned int offset, unsigned int dlength)
{
	const struct xtuser_rule *rule;
	struct nlmsghdr *msg;
	int err, ret;

	msg = xtnetlink_newpkt(sk, NFXTM_CHAIN_SPLICE);
	if (chain->name != NULL &&
	    !mnl_attr_put_strz_check(msg, sk->bufsize,
	                             NFXTA_CHAIN_NAME, chain->name))
		return NFXTE_EMSGSIZE;
	if (!mnl_attr_put_u32_check(msg, sk->bufsize, NFXTA_SPLICE_OFFSET,
	    (offset == XTNL_CHAIN_APPEND) ? UINT32_MAX : offset) ||
	    !mnl_attr_put_u32_check(msg, sk->bufsize, NFXTA_SPLICE_DLENGTH,
	    (dlength == XTNL_CHAIN_FLUSH) ? UINT32_MAX : dlength))
		return NFXTE_EMSGSIZE;

	ret = xtnetlink_send(sk, msg, msg->nlmsg_len);
	if (ret < 0)
		return ret;
	ret = xtnetlink_recv(sk, sk->buf, sk->bufsize);
	if (ret < 0)
		return ret;
	ret = xtnetlink_pkt_error(sk->buf, ret);
	if (ret != 0)
		return ret;

	HXlist_for_each_entry(rule, &chain->rule_list, anchor) {
		ret = xtnetlink_emit_rule(sk, rule);
		if (ret != 0)
			break;
	}

	err = ret;
	msg = xtnetlink_newpkt(sk, (err != 0) ? NFXTM_ABORT : NFXTM_COMMIT);
	ret = xtnetlink_send(sk, msg, msg->nlmsg_len);
	if (ret < 0)
		return (err != 0) ? err : ret;
	ret = xtnetlink_recv(sk, sk->buf, sk->bufsize);
	if (ret < 0)
		return (err != 0) ? err : ret;
	return (err != 0) ? err : xtnetlink_pkt_error(sk->buf, ret);
}

EXPORT_SYMBOL int
xtnetlink_chain_promote(struct xtnetlink_handle *sk,
                        const struct xtnetlink_promote_param *par)
{
	struct nlmsghdr *msg;
	int ret;

	msg = xtnetlink_newpkt(sk, NFXTM_CHAIN_PROMOTE);
	if (!mnl_attr_put_strz_check(msg, sk->bufsize,
	                             NFXTA_CHAIN_NAME, par->name) ||
	    !mnl_attr_put_u32_check(msg, sk->bufsize,
	                            NFXTA_HOOK_INDEX, par->hooknum) ||
	    !mnl_attr_put_u32_check(msg, sk->bufsize,
	                            NFXTA_HOOK_PRIORITY, par->priority) ||
	    !mnl_attr_put_u32_check(msg, sk->bufsize,
	                            NFXTA_HOOK_NFPROTO, par->nfproto))
		return NFXTE_EMSGSIZE;

	ret = xtnetlink_send(sk, msg, msg->nlmsg_len);
	if (ret < 0)
		return ret;
	ret = xtnetlink_recv(sk, sk->buf, sk->bufsize);
	if (ret < 0)
		return ret;
	return xtnetlink_pkt_error(sk->buf, ret);
}

EXPORT_SYMBOL int
xtnetlink_chain_demote(struct xtnetlink_handle *sk, const char *chain_name)
{
	struct nlmsghdr *msg;
	int ret;

	msg = xtnetlink_newpkt(sk, NFXTM_CHAIN_DEMOTE);
	if (!mnl_attr_put_strz_check(msg, sk->bufsize,
	                             NFXTA_CHAIN_NAME, chain_name))
		return NFXTE_EMSGSIZE;
	ret = xtnetlink_send(sk, msg, msg->nlmsg_len);
	if (ret < 0)
		return ret;
	ret = xtnetlink_recv(sk, sk->buf, sk->bufsize);
	if (ret < 0)
		return ret;
	return xtnetlink_pkt_error(sk->buf, ret);
}

/**
 * Loop over all chains in the given table and issue Netlink messages for
 * creating, populating and promoting them.
 */
static int xtnetlink_table_replace2(struct xtnetlink_handle *sk,
                                    const struct xtuser_table *table)
{
	struct xtuser_table_iter *iter;
	const struct xtuser_chain *chain;
	struct xtnetlink_promote_param pr;
	int ret = NFXTE_SUCCESS;

	/*
	 * First create all chains, so that jumps and gotos can
	 * properly resolve during chain splicing.
	 */
	iter = xtuser_table_travinit(const_cast1(struct xtuser_table *, table));
	if (iter == NULL)
		return -errno;
	while ((chain = xtuser_table_traverse(iter)) != NULL) {
		ret = xtnetlink_chain_create(sk, chain);
		if (ret != NFXTE_SUCCESS)
			goto out;
	}
	xtuser_table_travfree(iter);

	iter = xtuser_table_travinit(const_cast1(struct xtuser_table *, table));
	if (iter == NULL)
		return -errno;
	while ((chain = xtuser_table_traverse(iter)) != NULL) {
		ret = xtnetlink_chain_splice(sk, chain,
		      XTNL_CHAIN_APPEND, XTNL_CHAIN_FLUSH);
		if (ret != NFXTE_SUCCESS)
			break;

		if (chain->hook_nfproto == XTUSER_INVALID_NFPROTO)
			continue;
		pr.name     = chain->name;
		pr.hooknum  = chain->hook_num;
		pr.priority = chain->hook_priority;
		pr.nfproto  = chain->hook_nfproto;
		
		ret = xtnetlink_chain_promote(sk, &pr);
		if (ret != NFXTE_SUCCESS)
			break;
	}
 out:
	xtuser_table_travfree(iter);
	return ret;
}

/**
 * Issue a table replace request, using the table given by the caller.
 */
EXPORT_SYMBOL int xtnetlink_table_replace(struct xtnetlink_handle *sk,
                                          const struct xtuser_table *table)
{
	struct nlmsghdr *msg;
	int ret;

	msg = xtnetlink_newpkt(sk, NFXTM_TABLE_REPLACE);
	ret = xtnetlink_send(sk, msg, msg->nlmsg_len);
	if (ret < 0)
		return ret;
	ret = xtnetlink_recv(sk, sk->buf, sk->bufsize);
	if (ret < 0)
		return ret;
	ret = xtnetlink_pkt_error(sk->buf, ret);
	if (ret != NFXTE_SUCCESS)
		return ret;

	ret = xtnetlink_table_replace2(sk, table);
	if (ret != NFXTE_SUCCESS) {
		msg = xtnetlink_newpkt(sk, NFXTM_ABORT);
		if (xtnetlink_send(sk, msg, msg->nlmsg_len) < 0)
			return ret;
		xtnetlink_recv(sk, sk->buf, sk->bufsize);
		return ret;
	}

	msg = xtnetlink_newpkt(sk, NFXTM_COMMIT);
	ret = xtnetlink_send(sk, msg, msg->nlmsg_len);
	if (ret < 0)
		return ret;
	ret = xtnetlink_recv(sk, sk->buf, sk->bufsize);
	if (ret < 0)
		return ret;
	return xtnetlink_pkt_error(sk->buf, ret);
}

static int
xtnetlink_parse_mt(struct xtuser_rule *rule, const struct nlattr *cont,
                   enum xtuser_action_type type)
{
	const struct nlattr *attr;
	unsigned int attr_type, revision = UINT_MAX;
	const char *name = NULL;
	void *data = NULL;
	size_t dsize = 0;

	mnl_attr_for_each_nested(attr, cont) {
		attr_type = mnl_attr_get_type(attr);
		if (attr_type == NFXTA_ACTION_NAME) {
			name = mnl_attr_get_str(attr);
		} else if (attr_type == NFXTA_REVISION) {
			revision = mnl_attr_get_u32(attr);
		} else if (attr_type == NFXTA_ACTION_DATA) {
			dsize = mnl_attr_get_payload_len(attr);
			data  = mnl_attr_get_payload(attr);
		}
	}
	if (name == NULL || revision == UINT_MAX)
		return -EIO;
	return xtuser_rule_push_mt(rule, type, name, revision, data, dsize);
}

static int
xtnetlink_parse_jg(struct xtuser_rule *rule, const struct nlattr *attr,
                   enum xtuser_action_type type)
{
	return xtuser_rule_push_jg(rule, type, mnl_attr_get_str(attr));
}

/**
 * Handle a NFXTM_RULE_ENTRY message, which indicates a rule. Creates a rule
 * object and attaches it to the parent chain.
 */
static int xtnetlink_parse_rule(struct xtnetlink_dp_state *s, const void *buf)
{
	const struct nlmsghdr *msg = buf;
	const struct nlattr *attr;
	struct xtuser_rule *rule;
	int ret;

	rule = xtuser_rule_new();
	if (rule == NULL)
		return -errno;

	mnl_attr_for_each(attr, msg, sizeof(struct nfgenmsg)) {
		unsigned int type = mnl_attr_get_type(attr);

		if (type == NFXTA_VERDICT)
			ret = xtuser_rule_push_verdict(rule,
			      mnl_attr_get_u32(attr));
		else if (type == NFXTA_JUMP_CHAIN)
			ret = xtnetlink_parse_jg(rule, attr,
			      XTUSER_ACTION_JUMP);
		else if (type == NFXTA_GOTO_CHAIN)
			ret = xtnetlink_parse_jg(rule, attr,
			      XTUSER_ACTION_GOTO);
		else if (type == NFXTA_MATCH)
			ret = xtnetlink_parse_mt(rule, attr,
			      XTUSER_ACTION_MATCH);
		else if (type == NFXTA_TARGET)
			ret = xtnetlink_parse_mt(rule, attr,
			      XTUSER_ACTION_TARGET);
		else
			fprintf(stderr, "WARNING: unknown action attribute "
			        "type %u\n", type);
		if (ret != 0) {
			xtuser_rule_free(NULL, rule);
			return ret;
		}
	}
	xtuser_chain_push(s->chain, rule);
	return 0;
}

/**
 * Handle the NFXTM_CHAIN_DUMP message, which announces that a new chain starts
 * in the dump now.
 */
static int xtnetlink_parse_chain(struct xtnetlink_dp_state *s, const void *buf)
{
	const struct nlmsghdr *msg = buf;
	const struct nlattr *attr;
	unsigned int hook_nfproto = XTUSER_INVALID_NFPROTO;
	unsigned int hook_num = XTUSER_INVALID_HOOK;
	int hook_priority;
	const char *name;

	mnl_attr_for_each(attr, msg, sizeof(struct nfgenmsg)) {
		unsigned int type = mnl_attr_get_type(attr);

		if (type == NFXTA_CHAIN_NAME) {
			if (mnl_attr_get_payload_len(attr) >
			    sizeof(s->chain->name))
				fprintf(stderr, "WARNING: xt_nfnetlink sent a "
				        "chain name larger than anticipated "
				        "by this library. Truncating name.\n");
			name = mnl_attr_get_str(attr);
			continue;
		} else if (type == NFXTA_HOOK_INDEX) {
			hook_num = mnl_attr_get_u32(attr);
		} else if (type == NFXTA_HOOK_PRIORITY) {
			hook_priority = mnl_attr_get_u32(attr);
		} else if (type == NFXTA_HOOK_NFPROTO) {
			hook_nfproto = mnl_attr_get_u32(attr);
		}
	}

	s->chain = xtuser_chain_create(s->table, name);
	if (s->chain == NULL)
		return -errno;

	s->chain->hook_nfproto  = hook_nfproto;
	s->chain->hook_num      = hook_num;
	s->chain->hook_priority = hook_priority;
	return 0;
}

/**
 * Handle the NFXTM_TABLE_DUMP message, which (obviously) announces the start
 * of the table in the table dump.
 */
static int xtnetlink_parse_table(struct xtnetlink_dp_state *s, const void *buf)
{
	s->table = xtuser_table_new();
	if (s->table == NULL)
		return -errno;
	return 0;
}

static int xtnetlink_dump_common2(const struct nlmsghdr *msg, void *vstate)
{
	struct xtnetlink_dp_state *state = vstate;
	unsigned int type;
	int ret;

	type = NFNL_MSG_TYPE(msg->nlmsg_type);
	if (type == NFXTM_TABLE_DUMP) {
		ret = xtnetlink_parse_table(state, msg);
		if (ret != 0)
			return ret;
	} else if (type == NFXTM_CHAIN_DUMP) {
		ret = xtnetlink_parse_chain(state, msg);
		if (ret != 0)
			return ret;
	} else if (type == NFXTM_RULE_ENTRY) {
		ret = xtnetlink_parse_rule(state, msg);
		if (ret != 0)
			return ret;
	} else {
		fprintf(stderr, "WARNING: %s: Unknown message type %u\n",
		        __func__, type);
		return 0;
	}
	return 0;
}

/**
 * Main dump function. This sends the dump request (both chain and table dump
 * requests are similar), parses the results and fills in xtuser structures.
 */
static int xtnetlink_dump_common(struct xtnetlink_handle *sk,
                                 struct xtnetlink_dp_state *state)
{
	struct nlmsghdr *msg;
	int ret;

	msg = xtnetlink_newpkt(sk, state->op);
	if (state->op == NFXTM_CHAIN_DUMP &&
	    !mnl_attr_put_strz_check(msg, sk->bufsize,
	                             NFXTA_CHAIN_NAME, state->name))
		return NFXTE_EMSGSIZE;
	ret = xtnetlink_send(sk, msg, msg->nlmsg_len);
	if (ret < 0)
		return ret;

	return xtnetlink_mrecv(sk, xtnetlink_dump_common2, state);
}

/**
 * Dump kernel chain by name and store it in a xtuser_chain.
 */
EXPORT_SYMBOL struct xtuser_chain *
xtnetlink_chain_dump(struct xtnetlink_handle *sk, const char *name)
{
	struct xtnetlink_dp_state state;
	int ret;

	memset(&state, 0, sizeof(state));
	state.op   = NFXTM_CHAIN_DUMP;
	state.name = name;
	ret = xtnetlink_dump_common(sk, &state);
	if (ret != 0) {
		xtuser_chain_free(state.chain);
		errno = -ret;
		return NULL;
	}
	return state.chain;
}

/**
 * Retrieve the entire table from the kernel and provide it in a xtuser_table.
 */
EXPORT_SYMBOL struct xtuser_table *
xtnetlink_table_dump(struct xtnetlink_handle *sk)
{
	struct xtnetlink_dp_state state = {.op = NFXTM_TABLE_DUMP};
	int ret;

	ret = xtnetlink_dump_common(sk, &state);
	if (ret != 0) {
		xtuser_table_free(state.table);
		errno = -ret;
		return NULL;
	}
	return state.table;
}

static int xtnetlink_identify2(const struct nlmsghdr *nlmsg, void *data)
{
	unsigned int type = NFNL_MSG_TYPE(nlmsg->nlmsg_type);
	struct xtnetlink_identify *id = data;
	const struct nlattr *attr;

	if (type == NFXTM_IDENTIFY) {
		mnl_attr_for_each(attr, nlmsg, sizeof(struct nfgenmsg)) {
			unsigned int attr_type = mnl_attr_get_type(attr);

			if (attr_type == NFXTA_NAME) {
				id->name = HX_strdup(mnl_attr_get_str(attr));
				if (id->name == NULL)
					return -errno;
			} else if (attr_type == NFXTA_REVISION_MIN) {
				id->rev_min = mnl_attr_get_u32(attr);
			} else if (attr_type == NFXTA_REVISION_MAX) {
				id->rev_max = mnl_attr_get_u32(attr);
			}
		}
	}
	return 0;
}

EXPORT_SYMBOL int xtnetlink_identify(struct xtnetlink_handle *sk,
                                     struct xtnetlink_identify **idp)
{
	struct xtnetlink_identify *id;
	struct nlmsghdr *msg;
	ssize_t ret;

	msg = xtnetlink_newpkt(sk, NFXTM_IDENTIFY);
	ret = xtnetlink_send(sk, msg, msg->nlmsg_len);
	if (ret < 0)
		return ret;
	*idp = id = calloc(1, sizeof(*id));
	if (id == NULL)
		return -errno;
	id->rev_min = id->rev_max = -1;

	ret = xtnetlink_mrecv(sk, xtnetlink_identify2, id);
	if (ret != 0) {
		free(id);
		return ret;
	}
	return 0;
}

EXPORT_SYMBOL int
xtnetlink_ident_propget(const struct xtnetlink_identify *idp,
                        enum xtnetlink_ident_property prop, void *ptr)
{
	switch (prop) {
	case XTNETLINK_IDPROP_IMPNAME:
		*static_cast(const char **, ptr) = idp->name;
		break;
	case XTNETLINK_IDPROP_REVMIN:
		*static_cast(unsigned int *, ptr) = idp->rev_min;
		break;
	case XTNETLINK_IDPROP_REVMAX:
		*static_cast(unsigned int *, ptr) = idp->rev_max;
		break;
	default:
		return -EINVAL;
	}
	return 0;
}

EXPORT_SYMBOL void
xtnetlink_ident_free(struct xtnetlink_identify *id)
{
	free(id);
}

EXPORT_SYMBOL int xtnetlink_test_error(struct xtnetlink_handle *sk)
{
	struct nlmsghdr *msg;
	int ret;

	msg = xtnetlink_newpkt(sk, NFXTM_ERROR);
	ret = xtnetlink_send(sk, msg, msg->nlmsg_len);
	if (ret < 0)
		return ret;
	ret = xtnetlink_recv(sk, sk->buf, sk->bufsize);
	if (ret < 0)
		return ret;
	ret = xtnetlink_pkt_error(sk->buf, ret);
	printf("I expect an error. And it is: %s (%d)\n",
	       xtuser_strerror(ret), ret);

	msg = xtnetlink_newpkt(sk, NFXTM_ERROR);
	msg->nlmsg_flags |= NLM_F_DUMP;
	ret = xtnetlink_send(sk, msg, msg->nlmsg_len);
	if (ret < 0)
		return ret;

	return xtnetlink_mrecv(sk, NULL, NULL);
}
