/*
 *	Minimal test program for the chain splice operation
 *	written by Jan Engelhardt, 2012
 *
 *	This program is free software: you can redistribute it and/or modify
 *	it under the terms of the WTF Public License version 2 or
 *	(at your option) any later version.
 */
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <libnetfilter_xtables/props.h>
#include <libnetfilter_xtables/xtnetlink.h>
#include <libnetfilter_xtables/xtuser.h>
#include <libHX/misc.h>
#include <netinet/in.h>
#include <linux/netfilter.h>
#include "t-common.h"

static void tf_verdict_print(const struct xtuser_action *action)
{
	unsigned int v = xtuser_actionprop_verdict(action);

	if (v == NF_ACCEPT)
		printf(" -j ACCEPT\n");
	else if (v == NF_STOP)
		printf(" (+1 strange)\n");
	else if ((v & NF_VERDICT_MASK) == NF_DROP)
		printf(" -j DROP\n");
	else if ((v & NF_VERDICT_MASK) == NF_QUEUE)
		printf(" -j QUEUE --queue %u\n", v >> NF_VERDICT_QBITS);
	else
		printf(" (+1 wtf)\n");
}

static void tf_mt_print(const struct xtuser_action *action, char prefix)
{
	const void *data = xtuser_actionprop_data(action);
	size_t dsize = xtuser_actionprop_dsize(action);

	printf(" -%c %s.%u + <%zu bytes of data>\n",
	       prefix, xtuser_actionprop_name(action),
	       xtuser_actionprop_revision(action), dsize);
	if (t_hexdump)
		HX_hexdump(stdout, data, dsize);
}

static void tf_jg_print(const struct xtuser_action *action, char prefix)
{
	printf(" -%c %s\n", prefix, xtuser_actionprop_name(action));
}

static int tf_rule_print(const struct xtuser_rule *rule)
{
	struct xtuser_rule_iter *iter;
	const struct xtuser_action *action;
	unsigned int action_idx = 0;
	enum xtuser_action_type atype;

	iter = xtuser_rule_travinit(rule);
	while ((action = xtuser_rule_traverse(iter)) != NULL) {
		printf("\t\tAction #%u:", ++action_idx);
		atype = xtuser_actionprop_type(action);
		if (atype == XTUSER_ACTION_VERDICT)
			tf_verdict_print(action);
		else if (atype == XTUSER_ACTION_MATCH)
			tf_mt_print(action, 'm');
		else if (atype == XTUSER_ACTION_TARGET)
			tf_mt_print(action, 'j');
		else if (atype == XTUSER_ACTION_JUMP)
			tf_jg_print(action, 'j');
		else if (atype == XTUSER_ACTION_GOTO)
			tf_jg_print(action, 'g');
		else
			printf(" (+1 unknown)\n");
	}
	xtuser_rule_travfree(iter);
	return 0;
}

int tf_chain_print(const struct xtuser_chain *chain)
{
	struct xtuser_chain_iter *iter;
	const struct xtuser_rule *rule;
	unsigned int rule_idx = 0;
	unsigned int tmp;

	printf("Chain \"%s\"", xtuser_chainprop_name(chain));
	tmp = xtuser_chainprop_hooknfproto(chain);
	if (tmp != XTUSER_INVALID_NFPROTO)
		printf(" (nfproto %u, hook %u, priority %d)",
		       tmp, xtuser_chainprop_hooknum(chain),
		       xtuser_chainprop_hookprio(chain));
	printf("\n");

	iter = xtuser_chain_travinit(chain);
	while ((rule = xtuser_chain_traverse(iter)) != NULL) {
		printf("\tRule #%u\n", ++rule_idx);
		tf_rule_print(rule);
	}

	xtuser_chain_travfree(iter);
	printf("\n");
	return 0;
}

static int tf_chain_dump(int argc, const char **argv, struct test_param *par)
{
	struct xtuser_chain *chain;

	if (argc < 2) {
		fprintf(stderr, "A chain name is required\n");
		return EXIT_FAILURE;
	}
	chain = xtnetlink_chain_dump(par->xth, argv[1]);
	if (chain == NULL) {
		fprintf(stderr, "xtnetlink_chain_dump: %s\n",
		        xtuser_strerror(-errno));
		return EXIT_FAILURE;
	}

	tf_chain_print(chain);
	xtuser_chain_free(chain);
	return EXIT_SUCCESS;
}

const struct test_program tp_chain_dump = {
	.tfunc = tf_chain_dump,
};
