/*
 *	Minimal test program
 *	written by Jan Engelhardt, 2012
 *
 *	This program is free software: you can redistribute it and/or modify
 *	it under the terms of the WTF Public License version 2 or
 *	(at your option) any later version.
 */
#include <errno.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libHX/defs.h>
#include <libHX/init.h>
#include <libHX/option.h>
#include <libnetfilter_xtables/xtnetlink.h>
#include <libnetfilter_xtables/xtuser.h>
#include "t-common.h"

struct test_program_entry {
	const char *name;
	const struct test_program *prog;
};

static const struct test_program_entry t_subprogs[] = {
	{"abort",          &tp_abort},
	{"error",          &tp_error},
	{"identify",       &tp_identify},
	{"chain-create",   &tp_chain_create},
	{"chain-delete",   &tp_chain_delete},
	{"chain-demote",   &tp_chain_demote},
	{"chain-dump",     &tp_chain_dump},
#if 0
	{"chain-move",     &tp_chain_move},
#endif
	{"chain-promote",  &tp_chain_promote},
	{"chain-splice",   &tp_chain_splice},
	{"table-dump",     &tp_table_dump},
	{"table-flush",    &tp_table_flush},
	{"table-replace",  &tp_table_replace},
	{"table-replace2", &tp_table_replace2},
};

unsigned int t_hexdump;

static const struct HXoption t_common_options[] = {
	{.sh = 'x', .type = HXTYPE_NONE, .ptr = &t_hexdump,
	 .help = "Print hexdumps of all Netlink packet exchanges"},
	HXOPT_AUTOHELP,
	HXOPT_TABLEEND,
};

static const struct test_program_entry *
t_prog_lookup(const char *name)
{
	unsigned int i;

	for (i = 0; i < ARRAY_SIZE(t_subprogs); ++i)
		if (strcmp(t_subprogs[i].name, name) == 0)
			return &t_subprogs[i];
	return NULL;
}

static void t_prog_list(void)
{
	unsigned int i;

	printf("Available commands:\n");
	for (i = 0; i < ARRAY_SIZE(t_subprogs); ++i)
		printf("\t%s\n", t_subprogs[i].name);
}

static bool t_get_options(int *argc, const char ***argv,
    const struct HXoption *table)
{
	int ret;

	ret = HX_getopt(table, argc, argv, HXOPT_USAGEONERR);
	if (ret != HXOPT_ERR_SUCCESS)
		return false;
	if (t_hexdump)
		setenv("LIBNFXT_DEBUG", "1", true);
	return true;
}

int main(int argc, const char **argv)
{
	const struct test_program_entry *pent;
	const struct test_program *prog;
	struct test_param par;
	int ret;

	ret = HX_init();
	if (ret < 0) {
		fprintf(stderr, "HX_init: %s\n", strerror(-ret));
		return EXIT_FAILURE;
	}
	if (argc < 2) {
		fprintf(stderr, "Specify a command.\n");
		t_prog_list();
		goto out;
	}
	pent = t_prog_lookup(argv[1]);
	if (pent == NULL) {
		fprintf(stderr, "No such command.\n");
		t_prog_list();
		goto out;
	}
	prog = pent->prog;
	++argv;
	--argc;

	if (prog->options != NULL) {
		if (!t_get_options(&argc, &argv, prog->options))
			goto out;
	} else {
		if (!t_get_options(&argc, &argv, t_common_options))
			goto out;
	}

	ret = xtuser_init();
	if (ret < 0) {
		fprintf(stderr, "xtuser_init: %s\n", strerror(-ret));
		goto out;
	}
	par.xth = xtnetlink_open(XTNL_NO_FLAGS);
	if (par.xth == NULL) {
		perror("xtnetlink_open");
		ret = -errno;
		goto out2;
	}

	ret = prog->tfunc(argc, argv, &par);
	xtnetlink_close(par.xth);
 out2:
	xtuser_exit();
 out:
	HX_exit();
	return ret;
}
