/*
 *	Minimal test program for the chain splice operation
 *	with match/target data blocks
 *	written by Jan Engelhardt, 2012
 *
 *	This program is free software: you can redistribute it and/or modify
 *	it under the terms of the WTF Public License version 2 or
 *	(at your option) any later version.
 */
#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <libHX/option.h>
#include <libnetfilter_xtables/xtnetlink.h>
#include <libnetfilter_xtables/xtuser.h>
#include <netinet/in.h>
#include <linux/netfilter.h>
#include <linux/netfilter/xt_comment.h>
#include <linux/netfilter/xt_mark.h>
#include "t-common.h"

static unsigned int repl_promote;
static unsigned int repl_use_match, repl_use_target, repl_use_verdict;
static unsigned int repl_use_jump, repl_use_goto;

static const struct xt_mark_tginfo2
	repl_mark66 = {.mark = 0x66666666, .mask = 0xFFFFFFFF},
	repl_mark44 = {.mark = 0x44444444, .mask = 0xFFFFFFFF};

static int
make_basic_chain(struct xtuser_table *table, const char *name, const char *n2)
{
	struct xtuser_chain *chain;
	struct xtuser_rule *rule;
	int ret;

	chain = xtuser_chain_create(table, name);
	if (chain == NULL) {
		fprintf(stderr, "xtuser_chain_create: %s\n",
		        xtuser_strerror(-errno));
		return -errno;
	}
	rule = xtuser_rule_new();
	if (rule == NULL) {
		fprintf(stderr, "xtuser_rule_new: %s\n",
		        xtuser_strerror(-errno));
		return -errno;
	}
	if (repl_use_goto && n2 != NULL) {
		ret = xtuser_rule_push_goto(rule, n2);
		if (ret < 0)
			fprintf(stderr, "xtuser_rule_push_goto: %s\n",
			        xtuser_strerror(ret));
	}
	return xtuser_chain_push(chain, rule);
}

static struct xtuser_chain *
make_sample_chain(struct xtuser_table *table, const char *name,
                  const struct xt_mark_tginfo2 *mark_data)
{
	static const struct xt_comment_info comment_data =
		{.comment = "Xtables2 ni youkoso"};
	struct xtuser_chain *chain;
	struct xtuser_rule *rule;
	int ret;

	chain = xtuser_chain_create(table, name);
	if (chain == NULL) {
		fprintf(stderr, "xtuser_chain_create: %s\n",
		        xtuser_strerror(-errno));
		return NULL;
	}

	rule = xtuser_rule_new();
	if (rule == NULL) {
		fprintf(stderr, "xtuser_rule_new: %s\n",
		        xtuser_strerror(-errno));
		goto out;
	}

	if (repl_use_match) {
		ret = xtuser_rule_push_match(rule, "comment", 0, &comment_data,
		                             sizeof(comment_data));
		if (ret < 0)
			fprintf(stderr, "xtuser_rule_push_match: %s\n",
			        xtuser_strerror(ret));
	}
	if (repl_use_target) {
		ret = xtuser_rule_push_target(rule, "MARK", 2, mark_data,
		                              sizeof(mark_data));
		if (ret < 0)
			fprintf(stderr, "xtuser_rule_push_match: %s\n",
			        xtuser_strerror(ret));
	}
	if (repl_use_jump) {
		ret = xtuser_rule_push_jump(rule, "jumphere");
		if (ret < 0)
			fprintf(stderr, "xtuser_rule_push_jump: %s\n",
			        xtuser_strerror(ret));
	}
	if (repl_use_match) {
		ret = xtuser_rule_push_match(rule, "comment", 0, &comment_data,
		                             sizeof(comment_data));
		if (ret < 0)
			fprintf(stderr, "xtuser_rule_push_match: %s\n",
			        xtuser_strerror(ret));
	}
	if (repl_use_target) {
		ret = xtuser_rule_push_target(rule, "MARK", 2, mark_data,
	                              sizeof(mark_data));
		if (ret < 0)
			fprintf(stderr, "xtuser_rule_push_match: %s\n",
			        xtuser_strerror(ret));
	}
	if (repl_use_verdict) {
		ret = xtuser_rule_push_verdict(rule, NF_ACCEPT);
		if (ret < 0)
			fprintf(stderr, "xtuser_rule_push_match: %s\n",
			        xtuser_strerror(ret));
	}

	ret = xtuser_chain_push(chain, rule);
	if (ret < 0) {
		fprintf(stderr, "xtuser_chain_push: %s\n",
		        xtuser_strerror(ret));
		goto out;
	}
	return chain;

 out:
	xtuser_rule_free(NULL, rule);
	xtuser_chain_free(chain);
	return NULL;
}

static int
tf_table_replace(int argc, const char **argv, struct test_param *par)
{
	struct xtuser_table *table;
	struct xtuser_chain *chain;
	int ret;

	table = xtuser_table_new();
	if (table == NULL) {
		fprintf(stderr, "xtuser_table_new: %s\n",
		        xtuser_strerror(-errno));
		return EXIT_FAILURE;
	}

	ret = make_basic_chain(table, "gotohere", NULL);
	if (ret != 0)
		return EXIT_FAILURE;
	ret = make_basic_chain(table, "jumphere", "gotohere");
	if (ret != 0)
		return EXIT_FAILURE;

	chain = make_sample_chain(table, "foo6", &repl_mark66);
	if (chain == NULL)
		return EXIT_FAILURE;
	if (repl_promote) {
		xtuser_chain_propset(chain, XTUSER_CPROP_HOOKNUM,
		                     NF_INET_LOCAL_IN);
		xtuser_chain_propset(chain, XTUSER_CPROP_HOOKNFPROTO,
		                     NFPROTO_IPV6);
		xtuser_chain_propset(chain, XTUSER_CPROP_HOOKPRIO, 0);
	}
	chain = make_sample_chain(table, "foo4", &repl_mark44);
	if (chain == NULL)
		return EXIT_FAILURE;
	if (repl_promote) {
		xtuser_chain_propset(chain, XTUSER_CPROP_HOOKNUM,
		                     NF_INET_LOCAL_IN);
		xtuser_chain_propset(chain, XTUSER_CPROP_HOOKNFPROTO,
		                     NFPROTO_IPV4);
		xtuser_chain_propset(chain, XTUSER_CPROP_HOOKPRIO, 0);
	}

	ret = xtnetlink_table_replace(par->xth, table);
	printf("Replacing table... (%d) %s\n", ret, xtuser_strerror(ret));
	xtuser_table_free(table);
	return (ret == NFXTE_SUCCESS) ? EXIT_SUCCESS : EXIT_FAILURE;
}

static const struct HXoption tf_table_replace_options[] = {
	{.sh = 'P', .type = HXTYPE_NONE, .ptr = &repl_promote,
	 .help = "Utilize promote nomination"},
	{.sh = 'g', .type = HXTYPE_NONE, .ptr = &repl_use_goto,
	 .help = "Use some goto"},
	{.sh = 'j', .type = HXTYPE_NONE, .ptr = &repl_use_jump,
	 .help = "Use some jumps"},
	{.sh = 'm', .type = HXTYPE_NONE, .ptr = &repl_use_match,
	 .help = "Use some matches"},
	{.sh = 't', .type = HXTYPE_NONE, .ptr = &repl_use_target,
	 .help = "Use some targets"},
	{.sh = 'v', .type = HXTYPE_NONE, .ptr = &repl_use_verdict,
	 .help = "Use some verdicts"},
	{.sh = 'x', .type = HXTYPE_NONE, .ptr = &t_hexdump,
	 .help = "Print hexdumps of all Netlink packet exchanges"},
	HXOPT_AUTOHELP,
	HXOPT_TABLEEND,
};

const struct test_program tp_table_replace = {
	.options = tf_table_replace_options,
	.tfunc   = tf_table_replace,
};

static void
t_fiddle2(struct xtuser_table *table, const char *name, unsigned int nfproto)
{
	struct xtuser_chain *chain;
	struct xtuser_rule *rule;
	int ret;

	chain = xtuser_chain_create(table, name);
	assert(chain != NULL);

	rule = xtuser_rule_new();
	assert(rule != NULL);
	if (nfproto == NFPROTO_IPV6)
		ret = xtuser_rule_push_target(rule, "MARK", 2,
		      &repl_mark66, sizeof(repl_mark66));
	else if (nfproto == NFPROTO_IPV4)
		ret = xtuser_rule_push_target(rule, "MARK", 2,
		      &repl_mark44, sizeof(repl_mark44));
	else
		/* No mark for bridge! */
		ret = 0;
	assert(ret == 0);
	ret = xtuser_chain_push(chain, rule);
	assert(ret == 0);

	rule = xtuser_rule_new();
	assert(rule != NULL);
	ret = xtuser_rule_push_goto(rule, "common");
	assert(ret == 0);
	ret = xtuser_chain_push(chain, rule);
	assert(ret == 0);
	ret = xtuser_chain_propset(chain, XTUSER_CPROP_HOOKNUM, NF_INET_PRE_ROUTING);
	assert(ret == 0);
	ret = xtuser_chain_propset(chain, XTUSER_CPROP_HOOKNFPROTO, nfproto);
	assert(ret == 0);
	ret = xtuser_chain_propset(chain, XTUSER_CPROP_HOOKPRIO, 0);
	assert(ret == 0);
}

static int
tf_table_replace2(int argc, const char **argv, struct test_param *par)
{
	struct xtuser_table *table;
	struct xtuser_chain *chain;
	struct xtuser_rule *rule;
	int ret;

	table = xtuser_table_new();
	assert(table != NULL);

	t_fiddle2(table, "ipv6/filter/PREROUTING", NFPROTO_IPV6);
	t_fiddle2(table, "ipv4/filter/PREROUTING", NFPROTO_IPV4);
	t_fiddle2(table, "bridge/filter/PREROUTING", NFPROTO_BRIDGE);

	chain = xtuser_chain_create(table, "test1");
	assert(chain != NULL);
	rule = xtuser_rule_new();
	assert(rule != NULL);
	static const struct xt_mark_tginfo2 mk1 = {.mark = 0x01};
	ret = xtuser_rule_push_target(rule, "MARK", 2, &mk1, sizeof(mk1));
	assert(ret == 0);
	ret = xtuser_rule_push_goto(rule, "test2");
	assert(ret == 0);
	ret = xtuser_chain_push(chain, rule);
	assert(ret == 0);

	chain = xtuser_chain_create(table, "test2");
	rule = xtuser_rule_new();
	static const struct xt_mark_tginfo2 mk2 = {.mark = 0x02};
	ret = xtuser_rule_push_target(rule, "MARK", 2, &mk2, sizeof(mk2));
	assert(ret == 0);
	ret = xtuser_chain_push(chain, rule);
	assert(ret == 0);

	chain = xtuser_chain_create(table, "common");
	assert(chain != NULL);
	rule = xtuser_rule_new();
	assert(rule != NULL);
	static const struct xt_mark_tginfo2 mk3 = {.mask = 0xFF};
	ret = xtuser_rule_push_target(rule, "MARK", 2, &mk3, sizeof(mk3));
	assert(ret == 0);
	ret = xtuser_rule_push_jump(rule, "test1");
	assert(ret == 0);
	static const struct xt_mark_tginfo2 mk4 = {.mark = 0x04};
	ret = xtuser_rule_push_target(rule, "MARK", 2, &mk4, sizeof(mk4));
	assert(ret == 0);
	ret = xtuser_rule_push_verdict(rule, NF_ACCEPT);
	assert(ret == 0);
	ret = xtuser_rule_push_verdict(rule, NF_DROP);
	assert(ret == 0);
	ret = xtuser_chain_push(chain, rule);
	assert(ret == 0);
	rule = xtuser_rule_new();
	assert(rule != NULL);
	ret = xtuser_rule_push_verdict(rule, NF_DROP);
	assert(ret == 0);
	ret = xtuser_rule_push_verdict(rule, NF_ACCEPT);
	assert(ret == 0);

	ret = xtnetlink_table_replace(par->xth, table);
	printf("%s\n", xtuser_strerror(ret));
	assert(ret == 0);

	return EXIT_SUCCESS;
}

const struct test_program tp_table_replace2 = {
	.tfunc = tf_table_replace2,
};
