/*
 *	libnetfilter_xtables
 *	User representation - cached objects management
 *	Copyright © Jan Engelhardt, 2010-2012
 *
 *	This program is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation, either version 2 of the License, or
 *	(at your option) any later version.
 */
#include <errno.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libHX/defs.h>
#include <libHX/list.h>
#include <libHX/map.h>
#include <libHX/string.h>
#include <libnetfilter_xtables/xtuser.h>
#include "internal.h"

struct xtuser_chain_iter {
	const struct HXlist_head *rhead;
	const struct xtuser_chain *chain;
};

struct xtuser_rule_iter {
	const struct HXlist_head *ahead;
	const struct xtuser_rule *rule;
};

/**
 * quick foreach-traversal data block
 * @table:	table being walked
 * @call:	user-defined callback function
 * @priv:	user-defined private pointer
 */
struct xtuser_table_fedata {
	const struct xtuser_table *table;
	xtuser_table_travfunc_t call;
	void *priv;
};

EXPORT_SYMBOL struct xtuser_action *
xtuser_action_by_idx(const struct xtuser_rule *rule, unsigned int idx)
{
	struct HXlist_head *action;

	for (action = rule->action_list.next;
	     action != &rule->action_list.list;
	     action = action->next)
		if (idx-- == 0)
			return HXlist_entry(action, struct xtuser_action,
			       anchor);
	return NULL;
}

EXPORT_SYMBOL int
xtuser_action_propget(const struct xtuser_action *action,
                      enum xtuser_action_property prop, void *ptr)
{
	switch (prop) {
	case XTUSER_APROP_TYPE:
		*static_cast(enum xtuser_action_type *, ptr) = action->type;
		break;
	case XTUSER_APROP_VERDICT:
		*static_cast(unsigned int *, ptr) = action->verdict;
		break;
	case XTUSER_APROP_NAME:
		*static_cast(const char **, ptr) = action->mt.name;
		break;
	case XTUSER_APROP_REVISION:
		*static_cast(unsigned int *, ptr) = action->mt.revision;
		break;
	case XTUSER_APROP_DATA:
		*static_cast(void **, ptr) = action->mt.data;
		break;
	case XTUSER_APROP_DATASIZE:
		*static_cast(size_t *, ptr) = action->mt.dsize;
		break;
	default:
		return -EINVAL;
	}
	return 0;
}

/**
 * Free an action object. The caller needs to unlink it from
 * the action_list, if any.
 */
static void xtuser_action_free(struct xtuser_action *action)
{
	if (action->type == XTUSER_ACTION_MATCH ||
	    action->type == XTUSER_ACTION_TARGET) {
		free(action->mt.name);
		free(action->mt.data);
	} else if (action->type == XTUSER_ACTION_JUMP ||
	    action->type == XTUSER_ACTION_GOTO) {
		free(action->jg_target);
	}
	free(action);
}

/**
 * @chain:	chain to associate with; may be %NULL.
 *
 * Create a new blank rule without any matches/targets attached.
 */
EXPORT_SYMBOL struct xtuser_rule *xtuser_rule_new(void)
{
	struct xtuser_rule *rule;

	rule = malloc(sizeof(*rule));
	if (rule == NULL)
		return NULL;
	HXlist_init(&rule->anchor);
	HXclist_init(&rule->action_list);
	return rule;
}

/**
 * Destroys a rule.
 */
EXPORT_SYMBOL void
xtuser_rule_free(struct xtuser_chain *chain, struct xtuser_rule *rule)
{
	struct xtuser_action *action, *next;

	if (rule == NULL)
		return;
	if (chain != NULL && !HXlist_empty(&rule->anchor))
		HXclist_del(&chain->rule_list, &rule->anchor);
	HXlist_for_each_entry_safe(action, next, &rule->action_list, anchor)
		xtuser_action_free(action);
	free(rule);
}

EXPORT_SYMBOL struct xtuser_rule *
xtuser_rule_by_idx(const struct xtuser_chain *chain, unsigned int idx)
{
	struct HXlist_head *rule;

	for (rule = chain->rule_list.next; rule != &chain->rule_list.list;
	     rule = rule->next)
		if (idx-- == 0)
			return HXlist_entry(rule, struct xtuser_rule, anchor);
	return NULL;
}

EXPORT_SYMBOL struct xtuser_rule_iter *
xtuser_rule_travinit(const struct xtuser_rule *rule)
{
	struct xtuser_rule_iter *iter;

	iter = malloc(sizeof(*iter));
	if (iter == NULL)
		return NULL;

	iter->rule  = rule;
	iter->ahead = &rule->action_list.list;
	return iter;
}

/**
 * @trav:	traverser state object
 *
 * Returns the next object from the traverser. Adding and deleting rules from
 * the chain during traversal is supported.
 */
EXPORT_SYMBOL struct xtuser_action *
xtuser_rule_traverse(struct xtuser_rule_iter *iter)
{
	iter->ahead = iter->ahead->next;
	if (iter->ahead == &iter->rule->action_list.list) {
		errno = 0;
		return NULL;
	}
	return HXlist_entry(iter->ahead, struct xtuser_action, anchor);
}

/**
 * @trav:	traverser state object
 *
 * Frees the memory associated with a traverser object.
 */
EXPORT_SYMBOL void xtuser_rule_travfree(struct xtuser_rule_iter *iter)
{
	free(iter);
}

/**
 * @table:	local table to modify
 * @name:	name of the new chain
 *
 * Creates a new chain in the draft table. @table may be NULL, in which case
 * a homeless chain is created. When @name is NULL, an anonymous chain is
 * created (for use with xtuser_chain_splice).
 */
EXPORT_SYMBOL struct xtuser_chain *
xtuser_chain_create(struct xtuser_table *table, const char *name)
{
	struct xtuser_chain *chain;
	int ret;

	if (name != NULL && strlen(name) >= sizeof(chain->name)) {
		errno = ENAMETOOLONG;
		return NULL;
	}
	if (table != NULL && xtuser_chain_lookup(table, name) != NULL) {
		errno = EEXIST;
		return NULL;
	}
	chain = malloc(sizeof(*chain));
	if (chain == NULL)
		return NULL;
	strcpy(chain->name, name);
	HXclist_init(&chain->rule_list);
	chain->hook_nfproto = XTUSER_INVALID_NFPROTO;
	chain->hook_num     = XTUSER_INVALID_HOOK;
	if (table != NULL) {
		ret = HXmap_add(table->chains, chain->name, chain);
		if (ret < 0) {
			errno = -ret;
			free(chain);
			return NULL;
		}
	}
	return chain;
}

/**
 * Destroys the chain, rules, etc. This must only be called for chains that
 * are already unlinked from its table. If that is not the case - go use
 * xtuser_table_del instead.
 */
EXPORT_SYMBOL void xtuser_chain_free(struct xtuser_chain *chain)
{
	if (chain == NULL)
		/*
		 * Allows to use
		 * xtuser_chain_free(xtuser_chain_unlink(table, "foo")).
		 */
		return;
	xtuser_chain_flush(chain);
	free(chain);
}

EXPORT_SYMBOL void xtuser_chain_flush(struct xtuser_chain *chain)
{
	struct xtuser_rule *rule, *next;

	HXlist_for_each_entry_safe(rule, next, &chain->rule_list, anchor)
		xtuser_rule_free(chain, rule);
}

/**
 * @chain:	chain to walk
 * @cb:		user-defined callback function
 * @priv:	user-defined private pointer
 *
 * Traverse the chain and call the supplied callback function for each rule.
 * @cb may return false to stop the iteration.
 */
EXPORT_SYMBOL void xtuser_chain_foreach(const struct xtuser_chain *chain,
                                        xtuser_chain_travfunc_t cb, void *priv)
{
	const struct xtuser_rule *rule;

	HXlist_for_each_entry(rule, &chain->rule_list, anchor)
		if (!(*cb)(chain, rule, priv))
			break;
}

/**
 * @chain:	chain to walk
 *
 * The three functions xtuser_chain_{travinit,traverse,travfree} are used to
 * walk through a chain, but without using callbacks. There is also
 * xtuser_user_foreach that you can consider using instead.
 * This traverser does not cope with the deletion of the element it is
 * currently at.
 */
EXPORT_SYMBOL struct xtuser_chain_iter *
xtuser_chain_travinit(const struct xtuser_chain *chain)
{
	struct xtuser_chain_iter *iter;

	iter = malloc(sizeof(*iter));
	if (iter == NULL)
		return NULL;

	iter->chain = chain;
	iter->rhead = &chain->rule_list.list;
	return iter;
}

/**
 * @trav:	traverser state object
 *
 * Returns the next object from the traverser. Adding and deleting rules from
 * the chain during traversal is supported.
 */
EXPORT_SYMBOL struct xtuser_rule *
xtuser_chain_traverse(struct xtuser_chain_iter *iter)
{
	iter->rhead = iter->rhead->next;
	if (iter->rhead == &iter->chain->rule_list.list) {
		errno = 0;
		return NULL;
	}
	return HXlist_entry(iter->rhead, struct xtuser_rule, anchor);
}

/**
 * @trav:	traverser state object
 *
 * Frees the memory associated with a traverser object.
 */
EXPORT_SYMBOL void xtuser_chain_travfree(struct xtuser_chain_iter *iter)
{
	free(iter);
}

/**
 * Unlink a chain from a table. It may then be passed to xtuser_chain_destroy.
 */
EXPORT_SYMBOL struct xtuser_chain *
xtuser_chain_unlink(struct xtuser_table *table, const char *name)
{
	return HXmap_del(table->chains, name);
}

/**
 * @table:	table to search in
 * @name:	chain name to search for
 *
 * Search for a chain identified by @name in @table. Returns %NULL if not
 * found, or a pointer to the chain.
 */
EXPORT_SYMBOL struct xtuser_chain *
xtuser_chain_lookup(const struct xtuser_table *table, const char *name)
{
	return HXmap_get(table->chains, name);
}

EXPORT_SYMBOL int
xtuser_chain_propget(const struct xtuser_chain *chain,
                     enum xtuser_chain_property prop, void *ptr)
{
	switch (prop) {
	case XTUSER_CPROP_NAME:
		*static_cast(const char **, ptr) = chain->name;
		break;
	case XTUSER_CPROP_HOOKNUM:
		*static_cast(unsigned int *, ptr) = chain->hook_num;
		break;
	case XTUSER_CPROP_HOOKPRIO:
		*static_cast(int *, ptr) = chain->hook_priority;
		break;
	case XTUSER_CPROP_HOOKNFPROTO:
		*static_cast(uint8_t *, ptr) = chain->hook_nfproto;
		break;
	default:
		return -EINVAL;
	}
	return 0;
}

EXPORT_SYMBOL int
xtuser_chain_propset(struct xtuser_chain *chain,
                     enum xtuser_chain_property prop, ...)
{
	va_list argp;

	va_start(argp, prop);
	switch (prop) {
	case XTUSER_CPROP_HOOKNUM:
		chain->hook_num = va_arg(argp, unsigned int);
		break;
	case XTUSER_CPROP_HOOKPRIO:
		chain->hook_priority = va_arg(argp, int);
		break;
	case XTUSER_CPROP_HOOKNFPROTO:
		chain->hook_nfproto = va_arg(argp, unsigned int);
		break;
	default:
		va_end(argp);
		return -EINVAL;
	}
	va_end(argp);
	return 0;
}

EXPORT_SYMBOL int
xtuser_chain_push(struct xtuser_chain *chain, struct xtuser_rule *rule)
{
	HXclist_push(&chain->rule_list, &rule->anchor);
	return 0;
}

EXPORT_SYMBOL int
xtuser_rule_push_verdict(struct xtuser_rule *rule, unsigned int verdict)
{
	struct xtuser_action *action;

	action = malloc(sizeof(*action));
	if (action == NULL)
		return -errno;

	HXlist_init(&action->anchor);
	action->type = XTUSER_ACTION_VERDICT;
	action->verdict = verdict;
	HXclist_push(&rule->action_list, &action->anchor);
	return 0;
}

int xtuser_rule_push_mt(struct xtuser_rule *rule, enum xtuser_action_type type,
                        const char *name, unsigned int revision,
                        const void *data, size_t size)
{
	struct xtuser_action *action;
	int ret;

	action = malloc(sizeof(*action));
	if (action == NULL)
		return -errno;

	HXlist_init(&action->anchor);
	action->type = type;
	action->mt.name = HX_strdup(name);
	if (action->mt.name == NULL)
		goto out;
	action->mt.revision = revision;
	if (data == NULL || size == 0) {
		action->mt.data  = NULL;
		action->mt.dsize = 0;
	} else {
		action->mt.data = HX_memdup(data, size);
		if (action->mt.data == NULL)
			goto out;
		action->mt.dsize = size;
	}
	HXclist_push(&rule->action_list, &action->anchor);
	return 0;

 out:
	ret = -errno;
	if (action != NULL) {
		free(action->mt.name);
		free(action->mt.data);
	}
	free(action);
	return ret;
}

EXPORT_SYMBOL int
xtuser_rule_push_match(struct xtuser_rule *rule, const char *name,
                       unsigned int revision, const void *data, size_t size)
{
	return xtuser_rule_push_mt(rule, XTUSER_ACTION_MATCH, name,
	                           revision, data, size);
}

EXPORT_SYMBOL int
xtuser_rule_push_target(struct xtuser_rule *rule, const char *name,
                        unsigned int revision, const void *data, size_t size)
{
	return xtuser_rule_push_mt(rule, XTUSER_ACTION_TARGET, name,
	                           revision, data, size);
}

int xtuser_rule_push_jg(struct xtuser_rule *rule, enum xtuser_action_type type,
                        const char *chain)
{
	struct xtuser_action *action;
	int ret;

	action = malloc(sizeof(*action));
	if (action == NULL)
		return -errno;

	HXlist_init(&action->anchor);
	action->type = type;
	action->jg_target = HX_strdup(chain);
	if (action->jg_target == NULL) {
		ret = -errno;
		free(action);
		return ret;
	}
	HXclist_push(&rule->action_list, &action->anchor);
	return 0;
}

EXPORT_SYMBOL int
xtuser_rule_push_jump(struct xtuser_rule *rule, const char *chain)
{
	return xtuser_rule_push_jg(rule, XTUSER_ACTION_JUMP, chain);
}

EXPORT_SYMBOL int
xtuser_rule_push_goto(struct xtuser_rule *rule, const char *chain)
{
	return xtuser_rule_push_jg(rule, XTUSER_ACTION_GOTO, chain);
}

/**
 * Allocates an empty draft table. This can be used to construct a table
 * complete with rules before doing an atomic replace.
 *
 * On error, %NULL will be returned and errno is set.
 */
EXPORT_SYMBOL struct xtuser_table *xtuser_table_new(void)
{
	static const struct HXmap_ops ops = {
		.d_free = reinterpret_cast(void (*)(void *), xtuser_chain_free),
	};
	struct xtuser_table *table;

	table = malloc(sizeof(*table));
	if (table == NULL)
		return NULL;
	table->chains = HXmap_init5(HXMAPT_DEFAULT, HXMAP_SKEY, &ops, 0, 0);
	if (table->chains == NULL) {
		free(table);
		return NULL;
	}
	return table;
}

/**
 * @table:	table to face doom
 *
 * Kill a table and its subordinate chains, rules, etc.
 */
EXPORT_SYMBOL void xtuser_table_free(struct xtuser_table *table)
{
	if (table == NULL)
		return;
	HXmap_free(table->chains);
	free(table);
}

static bool xtuser_table_feproc(const struct HXmap_node *node, void *priv)
{
	const struct xtuser_table_fedata *fe = priv;

	return fe->call(fe->table, node->data, fe->priv);
}

/**
 * @table:	table to walk
 * @cb:		user-defined callback function
 * @priv:	user-defined private pointer
 *
 * Traverse the table and call the supplied callback function for each chain.
 * @cb may return false to stop the iteration. This is faster than
 * xtuser_table_traverse, but as a constraint, allows for no modification of
 * the table during traversal.
 */
EXPORT_SYMBOL void xtuser_table_foreach(const struct xtuser_table *table,
                                        xtuser_table_travfunc_t cb, void *priv)
{
	struct xtuser_table_fedata fe =
		{.table = table, .call = cb, .priv = priv};
	HXmap_qfe(table->chains, xtuser_table_feproc, &fe);
}

/**
 * @table:	table to walk
 *
 * The three functions xtuser_table_{travinit,traverse,travfree} are used to
 * walk through a table, but without using callbacks. It is usually somewhat
 * slower than the callback variant, because it has to save its state when
 * returning to the caller. (Please use xtuser_table_foreach, if possible.)
 */
EXPORT_SYMBOL struct xtuser_table_iter *
xtuser_table_travinit(const struct xtuser_table *table)
{
	/*
	 * Somewhat hide the fact that this is just a HXmap_trav.
	 * Discourage people from feeding the obtained traverse object
	 * to HXmap_traverse manually.
	 */
	return reinterpret_cast(struct xtuser_table_iter *,
	       HXmap_travinit(table->chains, HXMAP_DTRAV));
}

/**
 * @trav:	traverser state object
 *
 * Returns the next object from the traverser. NOTE: Adding and deleting
 * chains from the table during traversal is supported, but may cause
 * previously-returned pointers to be returned again.
 */
EXPORT_SYMBOL struct xtuser_chain *
xtuser_table_traverse(struct xtuser_table_iter *iter)
{
	const struct HXmap_node *node;

	node = HXmap_traverse(reinterpret_cast(struct HXmap_trav *, iter));
	if (node == NULL)
		return NULL;
	return node->data;
}

/**
 * @trav:	traverser state object
 *
 * Frees the memory associated with a traverser object.
 */
EXPORT_SYMBOL void xtuser_table_travfree(struct xtuser_table_iter *iter)
{
	HXmap_travfree(reinterpret_cast(struct HXmap_trav *, iter));
}
